// Copyright 2016  Jonas Me
// See the 'AUTHORS' file at the top-level directory for a full list of authors.
//
// This Source Code Form is subject to the terms of the Mozilla Public
// License, v. 2.0. If a copy of the MPL was not distributed with this
// file, You can obtain one at http://mozilla.org/MPL/2.0/.

import 'package:test/test.dart';

import 'package:fastchemail/src/parser.dart' as parser;
import 'package:fastchemail/fastchemail.dart';

final TABLE_OK = [
  r"!#$%&'*+-/=?^_`{|}~@example.com",
  "user+mailbox@example.com",
  "customer/department=shipping@example.com",
  r"$A12345@example.com",
  "!def!xyz%abc@example.com",
  "_somename@example.com",
  "a@example.com",
  "a@x.y",
  "abc.def@example.com",
  "abc-def@example.com",
  "123@example.com",
  "xn--abc@example.com",
  "abc@x.y.z",
  "abc@xyz-example.com",
  "abc@c--n.com",
  "abc@xn--hxajbheg2az3al.xn--jxalpdlp",
];

final TABLE_ERR = [
  ["@", NoLocalPartException],
  ["@example.com", NoLocalPartException],
  ["abc@", NoDomainPartException],
  ["abc", NoSignAtException],
  ["abc@def@example.com", TooAtException],
  [".abc@example.com", LocalStartPeriodException],
  ["abc.@example.com", LocalEndPeriodException],
  ["abc@.example.com", DomainStartPeriodException],
  ["abc@example.com.", DomainEndPeriodException],
  ["ab..cd@example.com", ConsecutivePeriodException],
  ["abc@example..com", ConsecutivePeriodException],
  ["a@example", NoPeriodDomainException],
  [r"ab\c@example.com", WrongCharLocalException, '\\'],
  [r'ab"c"def@example.com', WrongCharLocalException, '"'],
  ["abc def@example.com", WrongCharLocalException, ' '],
  ["(comment)abc@example.com", WrongCharLocalException, '('],
  ["abc@[255.255.255.255]", WrongCharDomainException, '['],
  ["abc@(example.com", WrongCharDomainException, '('],
  ["abc@x.y_y.z", WrongCharDomainException, '_'],
  ["abc@-example.com", WrongStartLabelException, '-'],
  ["abc@example-.com", WrongEndLabelException, '-'],
  ["abc@x.-y.z", WrongStartLabelException, '-'],
  ["abc@x.y-.z", WrongEndLabelException, '-'],
  ["abc@1example.com", WrongStartLabelException, '1'],
  ["abc@x.123", WrongStartLabelException, '1'],
  ["abcd€f@example.com", NonAsciiException, '€'],
  ["abc@exámple.com", NonAsciiException, 'á'],
  ["a\tbc@example.com", ControlCharException, 2],
  ["abc@\texample.com", ControlCharException, 5],
];

void main() {
  group("parseEmail", () {
    test("", () {
      for (var x in TABLE_OK) {
        expect(parseEmail(x), equals((null)));
      }
    });

    test("", () {
      for (var x in TABLE_ERR) {
        final err = x[1];
        expect(() => parseEmail(x[0]), throwsA(new isInstanceOf<err>()));

        if (x.length == 3) {
          try {
            parseEmail(x[0]);
          } on ControlCharException catch (e) {
            expect(e.position, x[2]);
          } catch (e) {
            expect(e.charCode, x[2].codeUnitAt(0));
          }
        }
      }
    });
  });

  group("isValidEmail", () {
    test("", () => expect(isValidEmail(TABLE_OK[0]), true));
    test("", () => expect(isValidEmail(TABLE_ERR[0][0]), false));
  });

  String localPart = "";
  for (var i = 0; i < parser.MAX_LOCAL_PART; i++) {
    localPart = "${localPart}a";
  }

  String label = "";
  for (var i = 0; i < parser.MAX_LABEL; i++) {
    label = "${label}x";
  }
  label = "${label}.";

  String allLabels = "";
  for (var i = 0; i < 3; i++) {
    allLabels = "${allLabels}${label}";
  }

  String lastLabel = "";
  for (var i = 0; i < parser.MAX_DOMAIN_PART - allLabels.length; i++) {
    lastLabel = "${lastLabel}y";
  }

  group("length", () {
    test("", () {
      expect(
          parseEmail("${localPart}@${allLabels}${lastLabel}"), equals((null)));
    });

    test("", () {
      expect(() => parseEmail("a${localPart}@${allLabels}${lastLabel}"),
          throwsA(new isInstanceOf<LocalTooLongException>()));
    });
    test("", () {
      expect(() => parseEmail("${localPart}@${allLabels}${lastLabel}z"),
          throwsA(new isInstanceOf<DomainTooLongException>()));
    });
    test("", () {
      expect(() => parseEmail("${localPart}@${label}x${lastLabel}"),
          throwsA(new isInstanceOf<LabelTooLongException>()));
    });
  });
}
